import {
  MutationDeleteActivityArgs,
  MutationDeleteActivityReplyArgs,
  MutationDeleteCustomListArgs,
  MutationDeleteMediaListEntryArgs,
  MutationDeleteReviewArgs,
  MutationDeleteThreadArgs,
  MutationDeleteThreadCommentArgs,
  MutationRateReviewArgs,
  MutationSaveActivityReplyArgs,
  MutationSaveListActivityArgs,
  MutationSaveMediaListEntryArgs,
  MutationSaveMessageActivityArgs,
  MutationSaveRecommendationArgs,
  MutationSaveReviewArgs,
  MutationSaveTextActivityArgs,
  MutationSaveThreadArgs,
  MutationSaveThreadCommentArgs,
  MutationToggleActivityPinArgs,
  MutationToggleActivitySubscriptionArgs,
  MutationToggleFavouriteArgs,
  MutationToggleFollowArgs,
  MutationToggleLikeArgs,
  MutationToggleLikeV2Args,
  MutationToggleThreadSubscriptionArgs,
  MutationUpdateFavouriteOrderArgs,
  MutationUpdateMediaListEntriesArgs,
  MutationUpdateUserArgs,
  QueryActivityArgs,
  QueryActivityReplyArgs,
  QueryAiringScheduleArgs,
  QueryCharacterArgs,
  QueryExternalLinkSourceCollectionArgs,
  QueryFollowerArgs,
  QueryFollowingArgs,
  QueryLikeArgs,
  QueryMarkdownArgs,
  QueryMediaArgs,
  QueryMediaListArgs,
  QueryMediaListCollectionArgs,
  QueryMediaTagCollectionArgs,
  QueryMediaTrendArgs,
  QueryNotificationArgs,
  QueryPageArgs,
  QueryRecommendationArgs,
  QueryReviewArgs,
  QueryStaffArgs,
  QueryStudioArgs,
  QueryThreadArgs,
  QueryThreadCommentArgs,
  QueryUserArgs,
} from "./types/generated/graphql.ts"
import { Fetch, Fields, Fn, UpdateOperation } from "./types/Anilist.ts"
import type { AtLeastOne } from "./types/AtLeastOne.ts"
import { Media } from "./queries/Media.ts"
import { MediaTrend } from "./queries/MediaTrend.ts"
import { AiringSchedule } from "./queries/AiringSchedule.ts"
import { Character } from "./queries/Character.ts"
import { Staff } from "./queries/Staff.ts"
import { MediaList } from "./queries/MediaList.ts"
import { MediaListCollection } from "./queries/MediaListCollection.ts"
import { MediaTag } from "./queries/MediaTag.ts"
import { User } from "./queries/User.ts"
import { Studio } from "./queries/Studio.ts"
import { Review } from "./queries/Review.ts"
import { ActivityReply } from "./queries/ActivityReply.ts"
import { Thread } from "./queries/Thread.ts"
import { ThreadComment } from "./queries/ThreadComment.ts"
import { Recommendation } from "./queries/Recommendation.ts"
import { AniChartUser } from "./queries/AniChartUser.ts"
import { Markdown } from "./queries/Markdown.ts"
import { SiteStatistics } from "./queries/SiteStatistics.ts"
import { ExternalLinkSourceCollection } from "./queries/ExternalLinkSourceCollection.ts"
import { AuthorizationError } from "./utils/AuthorizationError.ts"
import { ResponseError } from "./utils/ResponseError.ts"
import { Favourites } from "./queries/Favourites.ts"
import {
  ActivityUnion,
  ListActivity,
  MessageActivity,
  TextActivity,
} from "./queries/ActivityUnion.ts"
import { LikeableUnion } from "./queries/LikeableUnion.ts"
import { NotificationUnion } from "./queries/NotificationUnion.ts"
import { Page } from "./queries/Page.ts"

const rewriteVarValues = (
  variable: PropertyKey | unknown | (PropertyKey | unknown)[],
) => {
  switch (typeof variable) {
    case "object":
      return `[${variable}]`
    case "string":
      return variable === variable.toUpperCase() ? `${variable}` : `"${variable}"`
    default:
      return variable
  }
}
const tab = (n: number) => new Array(n + 1).fill("").join("\t")
const t = tab
const replaceLast = (str: string, pattern: string | RegExp, replacement: string) => {
  const match = typeof pattern === "string"
    ? pattern
    : (str.match(new RegExp(pattern.source, "g")) || []).slice(-1)[0]
  if (!match) return str
  const last = str.lastIndexOf(match)
  return last !== -1 ? `${str.slice(0, last)}${replacement}${str.slice(last + match.length)}` : str
}

const updateOperation: UpdateOperation = (
  { query, variables, fields, level, hasSubField, isUnion },
) => {
  if (!query) query = `${fields[level]}  {\n\t%${fields[level]}\n}`
  const convertedType: [string[], string[]] = [[], []]
  if (variables) {
    const variable = Object.entries(variables)
    for (let i = 0; i <= variable.length - 1; i++) {
      convertedType[0].push(`${variable[i][0]}: ${rewriteVarValues(variable[i][1])}`)
    }
  }
  query = replaceLast(
    query,
    `%${fields[level - 1]}`,
    `${isUnion ? "... on " : ""}${fields[level]}${
      variables ? ` (${convertedType[0].join(", ")})` : ""
    }${hasSubField ? `${` {\n${t(level + 1)}%${fields.at(-1)}\n${t(level)}}`}` : ""}\n${t(level)}%${
      fields[level - 1]
    }`,
  )
  return {
    get() {
      // return query
      return query!.replace(/(\n|,)?(\t+| )%\w+\b/g, "")
    },
    set({ subField, variables, hasSubField, level, isUnion }) {
      fields.splice(level, 0, subField)
      while (fields.length - 1 > level) fields.pop()

      return updateOperation({ query, variables, level, fields, hasSubField, isUnion })
    },
  }
}

const fetch: Fetch = async (init) => {
  let host:
    | "https://graphql.anilist.co"
    | "https://anilist.co/api/v2/oauth/token"

  const body = {}
  const headers = new Headers({
    "Content-Type": "application/json",
    "Accept": "application/json",
  })
  if ("code" in init) {
    const {
      client_id,
      client_secret,
      code,
      redirect_uri,
      grant_type = "authorization_code",
    } = init
    host = "https://anilist.co/api/v2/oauth/token"
    Object.assign(body, {
      grant_type,
      client_id,
      client_secret,
      redirect_uri,
      code,
    })
  } else {
    host = "https://graphql.anilist.co"
    const { query, token } = init
    Object.assign(body, { query })
    if (token) headers.set("Authorization", `Bearer ${token}`)
  }

  const res = await self.fetch(host, {
    method: "POST",
    headers,
    body: JSON.stringify(body),
  })

  if (res.status !== 200) {
    if ("code" in init) throw new AuthorizationError(await res.json())
    throw new ResponseError(await res.json())
  }
  return await res.json()
}

export const FuzzyDate = ({ query, level }: Fields<typeof updateOperation>) => ({
  withDay() {
    query[0] = query[0].set({ subField: "day", level })
    return this
  },
  withMonth() {
    query[0] = query[0].set({ subField: "month", level })
    return this
  },
  withYear() {
    query[0] = query[0].set({ subField: "year", level })
    return this
  },
})

export const Deleted = ({ query, level }: Fields<typeof updateOperation>) => ({
  /** If an item has been successfully deleted */
  withDeleted() {
    query[0] = query[0].set({ subField: "deleted", level })
    console.log("asd")
    return this
  },
})

export const Client = function (auth?: { token: string }) {
  let operation: ReturnType<typeof updateOperation>
  return {
    get auth() {
      return {
        getToken(init: {
          code: string
          client_id: string
          client_secret: string
          redirect_uri: string
        }) {
          return fetch(init)
        },
      }
    },
    get query() {
      operation = updateOperation({
        fields: ["query"],
        level: 0,
      })

      return {
        get raw() {
          return {
            get() {
              return operation.get()
            },
          }
        },
        fetch() {
          return fetch({ query: this.raw.get()! })
        },
        Page(args: AtLeastOne<QueryPageArgs>, fn: Fn<typeof Page>) {
          operation = operation.set({
            subField: "Page",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          let tmpQuery
          fn(Page({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]

          return this
        },
        /** Media query */
        Media(args: AtLeastOne<QueryMediaArgs>, fn?: Fn<typeof Media>) {
          operation = operation.set({
            subField: "Media",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(Media({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }

          return this
        },
        /** Media Trend query */
        MediaTrend(args: QueryMediaTrendArgs, fn?: Fn<typeof MediaTrend>) {
          operation = operation.set({
            subField: "MediaTrend",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(MediaTrend({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Airing schedule query */
        AiringSchedule(args: QueryAiringScheduleArgs, fn?: Fn<typeof AiringSchedule>) {
          operation = operation.set({
            subField: "AiringSchedule",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(AiringSchedule({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Character query */
        Character(args: QueryCharacterArgs, fn?: Fn<typeof Character>) {
          operation = operation.set({
            subField: "Character",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(Character({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Staff query */
        Staff(args: QueryStaffArgs, fn?: Fn<typeof Staff>) {
          operation = operation.set({
            subField: "Staff",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(Staff({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Media list query */
        MediaList(args: QueryMediaListArgs, fn?: Fn<typeof MediaList>) {
          operation = operation.set({
            subField: "MediaList",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(MediaList({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Media list collection query, provides list pre-grouped by status & custom lists. User ID and Media Type arguments required. */
        MediaListCollection(
          args: QueryMediaListCollectionArgs,
          fn?: Fn<typeof MediaListCollection>,
        ) {
          operation = operation.set({
            subField: "MediaListCollection",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(MediaListCollection({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Collection of all the possible media genres */
        GenreCollection() {
          operation = operation.set({ subField: "GenreCollection", level: 1 })
          return this
        },
        /** Collection of all the possible media tags */
        MediaTagCollection(args: QueryMediaTagCollectionArgs, fn?: Fn<typeof MediaTag>) {
          operation = operation.set({
            subField: "MediaTagCollection",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(MediaTag({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** User query */
        User(args: QueryUserArgs, fn?: Fn<typeof User>) {
          operation = operation.set({
            subField: "User",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(User({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Get the currently authenticated user */
        Viewer(fn: Fn<typeof User>) {
          operation = operation.set({
            subField: "Viewer",
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(User({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Notification query */
        Notification(args: QueryNotificationArgs, fn: Fn<typeof NotificationUnion>) {
          operation = operation.set({
            subField: "Notification",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          let tmpQuery
          fn(NotificationUnion({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]
          return this
        },
        /** Studio query */
        Studio(args: QueryStudioArgs, fn?: Fn<typeof Studio>) {
          operation = operation.set({
            subField: "Studio",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(Studio({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Review query */
        Review(args: QueryReviewArgs, fn?: Fn<typeof Review>) {
          operation = operation.set({
            subField: "Review",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(Review({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Activity query */
        Activity(args: QueryActivityArgs, fn: Fn<typeof ActivityUnion>) {
          operation = operation.set({
            subField: "Activity",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          let tmpQuery
          fn(ActivityUnion({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]

          return this
        },
        /** Activity reply query */
        ActivityReply(args: QueryActivityReplyArgs, fn?: Fn<typeof ActivityReply>) {
          operation = operation.set({
            subField: "ActivityReply",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(ActivityReply({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Follow query */
        Following(args: QueryFollowingArgs, fn?: Fn<typeof User>) {
          operation = operation.set({
            subField: "Following",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(User({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Follow query */
        Follower(args: QueryFollowerArgs, fn?: Fn<typeof User>) {
          operation = operation.set({
            subField: "Follower",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(User({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Thread query */
        Thread(args: QueryThreadArgs, fn?: Fn<typeof Thread>) {
          operation = operation.set({
            subField: "Thread",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(Thread({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Comment query */
        ThreadComment(args: QueryThreadCommentArgs, fn?: Fn<typeof ThreadComment>) {
          operation = operation.set({
            level: 1,
            subField: "ThreadComment",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(ThreadComment({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Recommendation query */
        Recommendation(args: QueryRecommendationArgs, fn?: Fn<typeof Recommendation>) {
          operation = operation.set({
            subField: "Recommendation",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(Recommendation({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Like query */
        Like(args: QueryLikeArgs, fn?: Fn<typeof User>) {
          operation = operation.set({
            subField: "Like",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(User({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Provide AniList markdown to be converted to html (Requires auth) */
        Markdown(args: QueryMarkdownArgs, fn: Fn<typeof Markdown>) {
          operation = operation.set({
            subField: "Markdown",
            hasSubField: true,
            variables: args,
            level: 1,
          })
          let tmpQuery
          fn(Markdown({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]
          return this
        },
        AniChartUser(fn: Fn<typeof AniChartUser>) {
          operation = operation.set({ subField: "AniChartUser", hasSubField: true, level: 1 })
          let tmpQuery
          fn(AniChartUser({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]

          return this
        },
        /** Site statistics query */
        SiteStatistics(fn?: Fn<typeof SiteStatistics>) {
          operation = operation.set({
            subField: "SiteStatistics",
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(SiteStatistics({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** ExternalLinkSource collection query */
        ExternalLinkSourceCollection(
          args: QueryExternalLinkSourceCollectionArgs,
          fn?: Fn<typeof ExternalLinkSourceCollection>,
        ) {
          operation = operation.set({
            subField: "ExternalLinkSourceCollection",
            variables: args,
            hasSubField: true,
            level: 1,
          })
          if (!fn) operation = operation.set({ subField: "id", level: 2 })
          else {
            let tmpQuery
            fn(ExternalLinkSourceCollection({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
      }
    },
    get mutation() {
      operation = updateOperation({
        fields: ["mutation"],
        level: 0,
      })

      return {
        get raw() {
          return {
            get() {
              return operation.get()
            },
          }
        },
        fetch() {
          return fetch({
            query: this.raw.get()!,
            token: auth!.token,
          })
        },
        /** Update current user options */
        UpdateUser(
          args: MutationUpdateUserArgs,
          fn?: Fn<typeof User>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "UpdateUser",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(User({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Create or update a media list entry */
        SaveMediaListEntry(
          args: MutationSaveMediaListEntryArgs,
          fn?: Fn<typeof MediaList>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "SaveMediaListEntry",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(MediaList({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Update multiple media list entries to the same values */
        UpdateMediaListEntries(
          args: MutationUpdateMediaListEntriesArgs,
          fn?: Fn<typeof MediaList>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "UpdateMediaListEntries",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(MediaList({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Delete a media list entry */
        DeleteMediaListEntry(args: MutationDeleteMediaListEntryArgs) {
          operation = operation.set({
            level: 1,
            subField: "DeleteMediaListEntry",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          Deleted({ query: tmpQuery = [operation], level: 2 }).withDeleted()
          operation = tmpQuery[0]
          return this
        },
        /** Delete a custom list and remove the list entries from it */
        DeleteCustomList(args: MutationDeleteCustomListArgs) {
          operation = operation.set({
            level: 1,
            subField: "DeleteCustomList",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          Deleted({ query: tmpQuery = [operation], level: 2 }).withDeleted()
          operation = tmpQuery[0]
          return this
        },
        /** Create or update text activity for the currently authenticated user */
        SaveTextActivity(args: MutationSaveTextActivityArgs, fn?: Fn<typeof TextActivity>) {
          operation = operation.set({
            level: 1,
            subField: "TextActivity",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          if (!fn) TextActivity({ query: tmpQuery = [operation], level: 2 }).withId()
          else fn(TextActivity({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]
          return this
        },
        /** Create or update message activity for the currently authenticated user */
        SaveMessageActivity(
          args: MutationSaveMessageActivityArgs,
          fn?: Fn<typeof MessageActivity>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "MessageActivity",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          if (!fn) MessageActivity({ query: tmpQuery = [operation], level: 2 }).withId()
          else fn(MessageActivity({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]
          return this
        },
        /** Update list activity (Mod Only) */
        SaveListActivity(args: MutationSaveListActivityArgs, fn?: Fn<typeof ListActivity>) {
          operation = operation.set({
            level: 1,
            subField: "ListActivity",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          if (!fn) ListActivity({ query: tmpQuery = [operation], level: 2 }).withId()
          else fn(ListActivity({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]
          return this
        },
        /** Delete an activity item of the authenticated users */
        DeleteActivity(args: MutationDeleteActivityArgs) {
          operation = operation.set({
            level: 1,
            subField: "DeleteActivity",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          Deleted({ query: tmpQuery = [operation], level: 2 }).withDeleted()
          operation = tmpQuery[0]
          return this
        },
        /** Toggle activity to be pinned to the top of the user's activity feed */
        ToggleActivityPin(args: MutationToggleActivityPinArgs, fn: Fn<typeof ActivityUnion>) {
          operation = operation.set({
            level: 1,
            subField: "ToggleActivityPin",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          fn(ActivityUnion({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]
          return this
        },
        /** Toggle the subscription of an activity item */
        ToggleActivitySubscription(
          args: MutationToggleActivitySubscriptionArgs,
          fn: Fn<typeof ActivityUnion>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "ToggleActivitySubscription",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          fn(ActivityUnion({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]
          return this
        },
        /** Create or update an activity reply */
        SaveActivityReply(
          args: MutationSaveActivityReplyArgs,
          fn?: Fn<typeof ActivityReply>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "SaveActivityReply",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(ActivityReply({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Delete an activity reply of the authenticated users */
        DeleteActivityReply(args: MutationDeleteActivityReplyArgs) {
          operation = operation.set({
            level: 1,
            subField: "DeleteActivityReply",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          Deleted({ query: tmpQuery = [operation], level: 2 }).withDeleted()
          operation = tmpQuery[0]
          return this
        },
        /** Add or remove a like from a likeable type. Returns all the users who liked the same model */
        ToggleLike(
          args: MutationToggleLikeArgs,
          fn?: Fn<typeof User>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "ToggleLike",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(User({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Add or remove a like from a likeable type. */
        ToggleLikeV2(args: MutationToggleLikeV2Args, fn: Fn<typeof LikeableUnion>) {
          operation = operation.set({
            level: 1,
            subField: "ToggleLikeV2",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          fn(LikeableUnion({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]
          return this
        },
        /** Toggle the un/following of a user */
        ToggleFollow(
          args: MutationToggleFollowArgs,
          fn?: Fn<typeof User>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "ToggleFollow",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(User({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Favourite or unfavourite an anime, manga, character, staff member, or studio */
        ToggleFavourite(args: MutationToggleFavouriteArgs, fn: Fn<typeof Favourites>) {
          operation = operation.set({
            level: 1,
            subField: "ToggleFavourite",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          fn(Favourites({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]
          return this
        },
        /** Update the order favourites are displayed in */
        UpdateFavouriteOrder(args: MutationUpdateFavouriteOrderArgs, fn: Fn<typeof Favourites>) {
          operation = operation.set({
            level: 1,
            subField: "UpdateFavouriteOrder",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          fn(Favourites({ query: tmpQuery = [operation], level: 2 }))
          operation = tmpQuery[0]

          return this
        },
        /** Create or update a review */
        SaveReview(
          args: MutationSaveReviewArgs,
          fn?: Fn<typeof Review>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "SaveReview",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(Review({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Delete a review */
        DeleteReview(args: MutationDeleteReviewArgs) {
          operation = operation.set({
            level: 1,
            subField: "DeleteReview",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          Deleted({ query: tmpQuery = [operation], level: 2 }).withDeleted()
          operation = tmpQuery[0]
          return this
        },
        /** Rate a review */
        RateReview(
          args: MutationRateReviewArgs,
          fn?: Fn<typeof Review>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "RateReview",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(Review({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Recommendation a media */
        SaveRecommendation(
          args: MutationSaveRecommendationArgs,
          fn?: Fn<typeof Recommendation>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "SaveRecommendation",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(Recommendation({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Create or update a forum thread */
        SaveThread(
          args: MutationSaveThreadArgs,
          fn?: Fn<typeof Thread>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "SaveThread",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(Thread({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Delete a thread */
        DeleteThread(args: MutationDeleteThreadArgs) {
          operation = operation.set({
            level: 1,
            subField: "DeleteThread",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          Deleted({ query: tmpQuery = [operation], level: 2 }).withDeleted()
          operation = tmpQuery[0]
          return this
        },
        /** Toggle the subscription of a forum thread */
        ToggleThreadSubscription(
          args: MutationToggleThreadSubscriptionArgs,
          fn?: Fn<typeof Thread>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "ToggleThreadSubscription",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(Thread({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Create or update a thread comment */
        SaveThreadComment(
          args: MutationSaveThreadCommentArgs,
          fn?: Fn<typeof ThreadComment>,
        ) {
          operation = operation.set({
            level: 1,
            subField: "SaveThreadComment",
            hasSubField: true,
            variables: args,
          })
          if (!fn) operation = operation.set({ level: 2, subField: "id" })
          else {
            let tmpQuery
            fn(ThreadComment({ query: tmpQuery = [operation], level: 2 }))
            operation = tmpQuery[0]
          }
          return this
        },
        /** Delete a thread comment */
        DeleteThreadComment(args: MutationDeleteThreadCommentArgs) {
          operation = operation.set({
            level: 1,
            subField: "DeleteThreadComment",
            hasSubField: true,
            variables: args,
          })
          let tmpQuery
          Deleted({ query: tmpQuery = [operation], level: 2 }).withDeleted()
          operation = tmpQuery[0]
          return this
        },
        UpdateAniChartSettings() {
          throw "To be Implemented"
        },
        UpdateAniChartHighlights() {
          throw "To be Implemented"
        },
      }
    },
  }
}
