import { Fields, Fn, UpdateOperation } from "../types/Anilist.ts"
import {
  FavouritesAnimeArgs,
  FavouritesCharactersArgs,
  FavouritesMangaArgs,
  FavouritesStaffArgs,
  FavouritesStudiosArgs,
} from "../types/generated/graphql.ts"
import { CharacterConnection } from "./Character.ts"
import { MediaConnection } from "./Media.ts"
import { StaffConnection } from "./Staff.ts"
import { StudioConnection } from "./Studio.ts"

export const Favourites = ({ query, level }: Fields<UpdateOperation>) => ({
  /** Favourite anime */
  withAnime(args: FavouritesAnimeArgs | undefined, fn: Fn<typeof MediaConnection>) {
    query[0] = query[0].set({ subField: "anime", level, variables: args })
    let tmpQuery
    fn(MediaConnection({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** Favourite manga */
  withManga(args: FavouritesMangaArgs | undefined, fn: Fn<typeof MediaConnection>) {
    query[0] = query[0].set({ subField: "manga", level, variables: args })
    let tmpQuery
    fn(MediaConnection({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** Favourite characters */
  withCharacters(args: FavouritesCharactersArgs | undefined, fn: Fn<typeof CharacterConnection>) {
    query[0] = query[0].set({ subField: "characters", level, variables: args })
    let tmpQuery
    fn(CharacterConnection({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** Favourite staff */
  withStaff(args: FavouritesStaffArgs | undefined, fn: Fn<typeof StaffConnection>) {
    query[0] = query[0].set({ subField: "staff", level, variables: args })
    let tmpQuery
    fn(StaffConnection({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  /** Favourite studios */
  withStudios(args: FavouritesStudiosArgs | undefined, fn: Fn<typeof StudioConnection>) {
    query[0] = query[0].set({ subField: "studios", level, variables: args })
    let tmpQuery
    fn(StudioConnection({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
})
