import { Fields, Fn, UpdateOperation } from "../types/Anilist.ts"
import {
  PageActivitiesArgs,
  PageActivityRepliesArgs,
  PageAiringSchedulesArgs,
  PageCharactersArgs,
  PageFollowersArgs,
  PageFollowingArgs,
  //   PageInfo,
  PageLikesArgs,
  PageMediaArgs,
  PageMediaListArgs,
  PageMediaTrendsArgs,
  PageNotificationsArgs,
  PageRecommendationsArgs,
  PageReviewsArgs,
  PageStaffArgs,
  PageStudiosArgs,
  PageThreadCommentsArgs,
  PageThreadsArgs,
  PageUsersArgs,
} from "../types/generated/graphql.ts"
import { ActivityReply } from "./ActivityReply.ts"
import { ActivityUnion } from "./ActivityUnion.ts"
import { AiringSchedule } from "./AiringSchedule.ts"
import { Character } from "./Character.ts"
import { Media } from "./Media.ts"
import { MediaList } from "./MediaList.ts"
import { MediaTrend } from "./MediaTrend.ts"
import { NotificationUnion } from "./NotificationUnion.ts"
import { Recommendation } from "./Recommendation.ts"
import { Review } from "./Review.ts"
import { Staff } from "./Staff.ts"
import { Studio } from "./Studio.ts"
import { Thread } from "./Thread.ts"
import { ThreadComment } from "./ThreadComment.ts"
import { User } from "./User.ts"

export const PageInfo = ({ query, level }: Fields<UpdateOperation>) => ({
  /** The total number of items. Note: This value is not guaranteed to be accurate, do not rely on this for logic */
  withTotal() {
    query[0] = query[0].set({ subField: "total", level })
    return this
  },
  /** The count on a page */
  withPerPage() {
    query[0] = query[0].set({ subField: "perPage", level })
    return this
  },
  /** The current page */
  withCurrentPage() {
    query[0] = query[0].set({ subField: "currentPage", level })
    return this
  },
  /** The last page */
  withLastPage() {
    query[0] = query[0].set({ subField: "lastPage", level })
    return this
  },
  /** If there is another page */
  withHasNextPage() {
    query[0] = query[0].set({ subField: "hasNextPage", level })
    return this
  },
})

export const Page = ({ query, level }: Fields<UpdateOperation>) => ({
  /** The pagination information */
  withPageInfo(fn: Fn<typeof PageInfo>) {
    query[0] = query[0].set({ subField: "pageInfo", level, hasSubField: true })
    let tmpQuery
    fn(PageInfo({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withUsers(args: PageUsersArgs, fn: Fn<typeof User>) {
    query[0] = query[0].set({ subField: "users", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(User({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withMedia(args: PageMediaArgs, fn: Fn<typeof Media>) {
    query[0] = query[0].set({ subField: "media", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(Media({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withCharacters(args: PageCharactersArgs, fn: Fn<typeof Character>) {
    query[0] = query[0].set({ subField: "characters", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(Character({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withStaff(args: PageStaffArgs, fn: Fn<typeof Staff>) {
    query[0] = query[0].set({ subField: "staff", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(Staff({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withStudios(args: PageStudiosArgs, fn: Fn<typeof Studio>) {
    query[0] = query[0].set({ subField: "studios", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(Studio({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withMediaList(args: PageMediaListArgs, fn: Fn<typeof MediaList>) {
    query[0] = query[0].set({ subField: "mediaList", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(MediaList({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withAiringSchedules(args: PageAiringSchedulesArgs, fn: Fn<typeof AiringSchedule>) {
    query[0] = query[0].set({
      subField: "airingSchedules",
      level,
      hasSubField: true,
      variables: args,
    })
    let tmpQuery
    fn(AiringSchedule({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withMediaTrends(args: PageMediaTrendsArgs, fn: Fn<typeof MediaTrend>) {
    query[0] = query[0].set({ subField: "mediaTrends", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(MediaTrend({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withNotifications(args: PageNotificationsArgs, fn: Fn<typeof NotificationUnion>) {
    query[0] = query[0].set({
      subField: "notifications",
      level,
      hasSubField: true,
      variables: args,
    })
    let tmpQuery
    fn(NotificationUnion({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withFollowers(args: PageFollowersArgs, fn: Fn<typeof User>) {
    query[0] = query[0].set({ subField: "followers", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(User({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withFollowing(args: PageFollowingArgs, fn: Fn<typeof User>) {
    query[0] = query[0].set({ subField: "following", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(User({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withActivities(args: PageActivitiesArgs, fn: Fn<typeof ActivityUnion>) {
    query[0] = query[0].set({ subField: "activities", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(ActivityUnion({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withActivityReplies(args: PageActivityRepliesArgs, fn: Fn<typeof ActivityReply>) {
    query[0] = query[0].set({
      subField: "activityReplies",
      level,
      hasSubField: true,
      variables: args,
    })
    let tmpQuery
    fn(ActivityReply({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withThreads(args: PageThreadsArgs, fn: Fn<typeof Thread>) {
    query[0] = query[0].set({ subField: "threads", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(Thread({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withThreadComments(args: PageThreadCommentsArgs, fn: Fn<typeof ThreadComment>) {
    query[0] = query[0].set({
      subField: "threadComments",
      level,
      hasSubField: true,
      variables: args,
    })
    let tmpQuery
    fn(ThreadComment({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withReviews(args: PageReviewsArgs, fn: Fn<typeof Review>) {
    query[0] = query[0].set({ subField: "reviews", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(Review({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withRecommendations(args: PageRecommendationsArgs, fn: Fn<typeof Recommendation>) {
    query[0] = query[0].set({
      subField: "recommendations",
      level,
      hasSubField: true,
      variables: args,
    })
    let tmpQuery
    fn(Recommendation({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
  withLikes(args: PageLikesArgs, fn: Fn<typeof User>) {
    query[0] = query[0].set({ subField: "likes", level, hasSubField: true, variables: args })
    let tmpQuery
    fn(User({ query: tmpQuery = [query[0]], level: level + 1 }))
    query[0] = tmpQuery[0]
    return this
  },
})
